import logging
from typing import Optional

from pyspark.sql import DataFrame

from aporia.core.logging_utils import LOGGER_NAME
from aporia.core.types.field import FieldCategory
from aporia.pyspark.pyspark_training import calculate_dataframe_training_data
from aporia.pyspark.pyspark_training_validator import PySparkTrainingValidator
from aporia.training.training_model import TrainingModel
from aporia.training.training_validator import TrainingDataset

logger = logging.getLogger(LOGGER_NAME)


class PySparkTrainingModel:
    """PySpark training model."""

    def __init__(self, training_model: TrainingModel):
        """Initializes a PySparkTrainingModel.

        Args:
            training_model: A training model object, which will be used to report training data.
        """
        self._training_model = training_model

    def log_pyspark_training_set(
        self, features: DataFrame, labels: DataFrame, raw_inputs: Optional[DataFrame] = None
    ):
        """See aporia.model.Model."""
        logger.debug("Logging pyspark training set.")
        with self._training_model.handle_error("Logging pyspark training set failed, error: {}"):
            version_schema = self._training_model.get_version_schema()
            validator = PySparkTrainingValidator(
                dataset_type=TrainingDataset.TRAINING, schema=version_schema
            )
            validator.validate_pyspark(features=features, labels=labels, raw_inputs=raw_inputs)

            raw_inputs_training_data = None
            if raw_inputs is not None:
                raw_inputs_training_data = calculate_dataframe_training_data(
                    raw_inputs, version_schema[FieldCategory.RAW_INPUTS]
                )

            self._training_model.log_training_set_aggregations(
                features=calculate_dataframe_training_data(
                    features, version_schema[FieldCategory.FEATURES]
                ),
                labels=calculate_dataframe_training_data(
                    labels, version_schema[FieldCategory.PREDICTIONS]
                ),
                raw_inputs=raw_inputs_training_data,
            )

    def log_pyspark_test_set(
        self,
        features: DataFrame,
        predictions: DataFrame,
        labels: DataFrame,
        raw_inputs: Optional[DataFrame] = None,
    ):
        """See aporia.model.Model."""
        logger.debug("Logging pyspark test set.")
        with self._training_model.handle_error("Logging test set failed, error: {}"):
            version_schema = self._training_model.get_version_schema()
            validator = PySparkTrainingValidator(
                dataset_type=TrainingDataset.TEST, schema=version_schema
            )
            validator.validate_pyspark(
                features=features, labels=labels, predictions=predictions, raw_inputs=raw_inputs
            )

            raw_inputs_training_data = None
            if raw_inputs is not None:
                raw_inputs_training_data = calculate_dataframe_training_data(
                    raw_inputs, version_schema[FieldCategory.RAW_INPUTS]
                )

            self._training_model.log_test_set_aggregations(
                features=calculate_dataframe_training_data(
                    features, version_schema[FieldCategory.FEATURES]
                ),
                predictions=calculate_dataframe_training_data(
                    predictions, version_schema[FieldCategory.PREDICTIONS]
                ),
                labels=calculate_dataframe_training_data(
                    labels, version_schema[FieldCategory.PREDICTIONS]
                ),
                raw_inputs=raw_inputs_training_data,
            )



import logging
from typing import Dict, List, Optional

from aporia.core.http_client import HttpClient
from aporia.core.logging_utils import LOGGER_NAME
from aporia.core.types.field import FieldType

VALID_FIELD_TYPES = [field_type.value for field_type in FieldType]

logger = logging.getLogger(LOGGER_NAME)


async def run_create_model_version_query(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    model_type: str,
    features: Dict[str, str],
    predictions: Dict[str, str],
    raw_inputs: Optional[Dict[str, str]] = None,
    metrics: Optional[Dict[str, str]] = None,
    model_data_type: Optional[str] = None,
    labels: Optional[List[str]] = None,
    feature_importance: Optional[Dict[str, float]] = None,
    mapping: Optional[Dict[str, str]] = None,
):
    """Defines the schema for a specific model version.

    Args:
        http_client: Http client
        model_id: Model ID
        model_version: Model Version
        model_type: Model type
        features: Feature fields
        predictions: Prediction fields
        raw_inputs: Raw input fields.
        metrics: Prediction metric fields.
        model_data_type: Model data type.
        labels: Labels of multi-label, multiclass or binary model.
        feature_importance: Features' importance.
        mapping: General mapping. See Notes.

    Notes:
        Allowed mapping fields are:
            * batch_id_column_name: The name of the key in the `raw_inputs`
                dict that holds the value of the `batch_id`.
            * relevance_column_name: The name of the key in the `predictions`
                dict that holds the value of the `relevance` score of models of type `ranking`.
            * actual_relevance_column_name: The name of the key in the `actuals`
                dict that holds the value of the `relevance` score of models of type `ranking`.
    """
    model_version_input = {
        "name": model_version,
        "model_type": model_type,
        "version_schema": {
            "features": features,
            "predictions": predictions,
            "raw_inputs": raw_inputs,
            "metrics": metrics,
            "feature_positions": generate_feature_positions(features),
        },
        "model_data_type": model_data_type,
        "labels": labels,
        "feature_importance": feature_importance,
        "mapping": mapping,
    }

    await http_client.post(
        url=f"/models/{model_id}/versions",
        data=model_version_input,
    )


def generate_feature_positions(features: Dict[str, str]) -> Dict[str, int]:
    """Generate features' position from a features dict.

    Args:
        features: features dict

    Returns:
        Features to positions dictionary.
    """
    return {field_name: index for index, field_name in enumerate(features.keys())}



import logging
from typing import List

from aporia.core.http_client import HttpClient
from aporia.core.logging_utils import LOGGER_NAME

logger = logging.getLogger(LOGGER_NAME)


async def log_inference_fragments(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    environment: str,
    serialized_fragments: List[dict],
    await_insert: bool,
):
    """Reports a batch of inference fragments.

    Args:
        http_client: Http client
        model_id: Model ID
        model_version: Model version
        environment: Environment in which aporia is running.
        serialized_fragments: List of serialized fragment dicts.
        await_insert: True if the server should wait for the fragments
            to be stored before responding to the sdk.
    """
    query = """
        mutation LogPredict(
            $modelId: String!,
            $modelVersion: String!,
            $environment: String!,
            $predictions: [Prediction]!
            $isSync: Boolean!
        ) {
            logPredictions(
                modelId: $modelId,
                modelVersion: $modelVersion,
                environment: $environment,
                predictions: $predictions
                isSync: $isSync
            ) {
                warnings
            }
        }
    """

    variables = {
        "modelId": model_id,
        "modelVersion": model_version,
        "environment": environment,
        "predictions": serialized_fragments,
        "isSync": await_insert,
    }

    result = await http_client.graphql(query, variables)
    for warning in result["logPredictions"]["warnings"]:
        logger.warning(warning)



from enum import Enum


class ModelColor(Enum):
    """Model colors."""

    BLUE = "blue"
    ARCTIC_BLUE = "arctic_blue"
    GREEN = "green"
    TURQUOISE = "turquoise"
    PINK = "pink"
    PURPLE = "purple"
    YELLOW = "yellow"
    RED = "red"


class ModelIcon(Enum):
    """Model Icons."""

    GENERAL = "general"
    CHURN_AND_RETENTION = "churn-and-retention"
    CONVERSION_PREDICT = "conversion-predict"
    ANOMALY = "anomaly"
    DYNAMIC_PRICING = "dynamic-pricing"
    EMAIL_FILTERING = "email-filtering"
    DEMAND_FORECASTING = "demand-forecasting"
    LTV = "ltv"
    PERSONALIZATION = "personalization"
    FRAUD_DETECTION = "fraud-detection"
    CREDIT_RISK = "credit-risk"
    RECOMMENDATIONS = "recommendations"



from itertools import repeat
from typing import Iterable, Optional

from pyspark.sql import DataFrame
from pyspark.sql.types import Row

from aporia.core.errors import AporiaError
from aporia.inference.inference_model import InferenceModel


class PySparkPrediction:
    """PySpark prediction representation."""

    def __init__(
        self,
        id: Row,
        features: Optional[Row] = None,
        predictions: Optional[Row] = None,
        raw_inputs: Optional[Row] = None,
        actuals: Optional[Row] = None,
    ):
        """Initializes a PySparkPrediction.

        Args:
            id: Prediction ID.
            features: Prediction features.
            predictions: Prediction results.
            raw_inputs: Prediction raw inputs.
            actuals: Prediction actual results.
        """
        self.id = id[0]
        self.features = None
        if features is not None:
            self.features = features.asDict()

        self.predictions = None
        if predictions is not None:
            self.predictions = predictions.asDict()

        self.raw_inputs = None
        if raw_inputs is not None:
            self.raw_inputs = raw_inputs.asDict()

        self.actuals = None
        if actuals is not None:
            self.actuals = actuals.asDict()


class PySparkInferenceModel:
    """Model object for reporting PySpark dataframes."""

    def __init__(self, inference_model: InferenceModel):
        """Initliazes a PySparkInferenceModel.

        Args:
            inference_model: An inference model object, which will be used to
                report the actual predictions.
        """
        self._inference_model = inference_model

    def log_batch_pyspark_raw_inputs(self, ids: DataFrame, raw_inputs: DataFrame):
        """See aporia.model.Model."""
        with self._inference_model.handle_error(
            "Logging PySpark raw_inputs batch failed, error: {}"
        ):
            self._validate_pyspark_dataframes(ids=ids, dataframes=[raw_inputs])

            for prediction in self._iter_pyspark_prediction_dataframes(
                ids=ids, raw_inputs=raw_inputs
            ):
                self._inference_model.log_raw_inputs(
                    id=prediction.id, raw_inputs=prediction.raw_inputs
                )

    def log_batch_pyspark_actuals(self, ids: DataFrame, actuals: DataFrame):
        """See aporia.model.Model."""
        with self._inference_model.handle_error("Logging PySpark actuals batch failed, error: {}"):
            self._validate_pyspark_dataframes(ids=ids, dataframes=[actuals])

            for prediction in self._iter_pyspark_prediction_dataframes(ids=ids, actuals=actuals):
                self._inference_model.log_actuals(id=prediction.id, actuals=prediction.actuals)

    def log_batch_pyspark_prediction(
        self,
        ids: DataFrame,
        features: DataFrame,
        predictions: DataFrame,
        raw_inputs: Optional[DataFrame] = None,
        actuals: Optional[DataFrame] = None,
    ):
        """See aporia.model.Model."""
        with self._inference_model.handle_error(
            "Logging PySpark prediction batch failed, error: {}"
        ):
            self._validate_pyspark_dataframes(
                ids=ids, dataframes=[features, predictions, raw_inputs, actuals]
            )

            for prediction in self._iter_pyspark_prediction_dataframes(
                ids=ids,
                features=features,
                predictions=predictions,
                raw_inputs=raw_inputs,
                actuals=actuals,
            ):
                self._inference_model.log_prediction(
                    id=prediction.id,
                    features=prediction.features,
                    predictions=prediction.predictions,
                    raw_inputs=prediction.raw_inputs,
                    actuals=prediction.actuals,
                )

    @staticmethod
    def _validate_pyspark_dataframes(
        ids: DataFrame,
        dataframes: Iterable[Optional[DataFrame]],
    ):
        expected_count = ids.count()
        for dataframe in dataframes:
            if dataframe is None:
                continue

            if dataframe.count() != expected_count:
                raise AporiaError("PySpark dataframe sizes do not match.")

        if len(ids.columns) != 1:
            raise AporiaError("Ids dataframe must contain exactly one column.")

    @staticmethod
    def _iter_pyspark_prediction_dataframes(
        ids: DataFrame,
        features: Optional[DataFrame] = None,
        predictions: Optional[DataFrame] = None,
        raw_inputs: Optional[DataFrame] = None,
        actuals: Optional[DataFrame] = None,
    ) -> Iterable[PySparkPrediction]:

        features_iter = repeat(None)
        if features is not None:
            features_iter = features.toLocalIterator()  # type: ignore

        predictions_iter = repeat(None)
        if predictions is not None:
            predictions_iter = predictions.toLocalIterator()  # type: ignore

        raw_inputs_iter = repeat(None)
        if raw_inputs is not None:
            raw_inputs_iter = raw_inputs.toLocalIterator()  # type: ignore

        actuals_iter = repeat(None)
        if actuals is not None:
            actuals_iter = actuals.toLocalIterator()  # type: ignore

        for id_row, features_row, predictions_row, raw_inputs_row, actuals_row in zip(
            ids.toLocalIterator(),
            features_iter,
            predictions_iter,
            raw_inputs_iter,
            actuals_iter,
        ):
            yield PySparkPrediction(
                id=id_row,
                features=features_row,
                predictions=predictions_row,
                raw_inputs=raw_inputs_row,
                actuals=actuals_row,
            )



from typing import Dict, Optional

from .context import get_context
from .core_api import safe_api_function


@safe_api_function("Creating model tags failed, error: {}")
def add_model_tags(model_id: str, tags: Dict[str, str]):
    """Adds or updates tags to an existing model.

    Each tag is a key:value pair, key and value must be strings.

    Args:
        model_id: Model ID
        tags: A mapping of tag keys to tag values

    Notes:
        * Each model is restricted to 10 tags
        * Tag keys are always converted to lowercase
        * If the `tags` parameter contains tag keys that were already
          defined for the model, their values will be updated.
    """
    context = get_context()
    context.event_loop.run_coroutine(
        context.http_client.post(url=f"/models/{model_id}/tags", data={"tags": tags})
    )


@safe_api_function("Deleting model tag failed, error: {}")
def delete_model_tag(model_id: str, tag_key: str):
    """Deletes a model tag.

    Args:
        model_id: Model ID
        tag_key: Tag key to delete

    Notes:
        * This function is best-effort, it will not fail if the tag doesn't exist.
    """
    context = get_context()
    context.event_loop.run_coroutine(
        context.http_client.delete(url=f"/models/{model_id}/tags/{tag_key}")
    )


@safe_api_function("Fetching model tags failed, error: {}")
def get_model_tags(model_id: str) -> Optional[Dict[str, str]]:
    """Fetches the tag keys and values of a model.

    Args:
        model_id: Model ID

    Returns:
        A dict mapping tag keys to values
    """
    context = get_context()
    return context.event_loop.run_coroutine(context.http_client.get(url=f"/models/{model_id}/tags"))



import logging
from typing import Optional

from pyspark.sql import DataFrame

from aporia.core.errors import AporiaError
from aporia.core.logging_utils import LOGGER_NAME
from aporia.core.types.field import FieldType
from aporia.training.training_validator import TrainingDataType, TrainingValidator
from .pyspark_utils import infer_schema_from_pyspark_dataframe

logger = logging.getLogger(LOGGER_NAME)


class PySparkTrainingValidator(TrainingValidator):
    """PySpark training data validator."""

    def validate_pyspark(
        self,
        features: DataFrame,
        labels: DataFrame,
        predictions: Optional[DataFrame] = None,
        raw_inputs: Optional[DataFrame] = None,
    ):
        """Validates pyspark training dataframes.

        Args:
            features: Features dataframe
            labels: Labels (predictions) dataframe
            predictions: Predictions dataframe
            raw_inputs: Raw inputs dataframe
        """
        for training_data_type, data in [
            (TrainingDataType.FEATURES, features),
            (TrainingDataType.LABELS, labels),
            (TrainingDataType.PREDICTIONS, predictions),
            (TrainingDataType.RAW_INPUTS, raw_inputs),
        ]:
            if data is None:
                continue

            self.validate_field_category_is_in_schema(training_data_type)
            self.validate_no_field_mismatch_in_category(
                columns=data.columns, training_data_type=training_data_type
            )

            self._validate_pyspark_field_types_match_schema(
                training_data_type=training_data_type, data=data
            )

    def _validate_pyspark_field_types_match_schema(
        self, training_data_type: TrainingDataType, data: DataFrame
    ):
        inferred_schema = infer_schema_from_pyspark_dataframe(data)
        if inferred_schema is None:
            raise AporiaError(
                "Validating {} dataframe failed - could not infer the schema of the {} dataframe".format(
                    self.dataset_type.value, training_data_type.value
                )
            )

        field_category = training_data_type.to_field_category()
        missing_fields = self.schema[field_category].keys() - set(data.columns)

        for field_name, field_type in self.schema[field_category].items():
            if field_name in missing_fields:
                continue

            inferred_field_type = FieldType(inferred_schema[field_name])
            # Don't verify boolean fields, since we're going to cast them anyway
            if field_type != FieldType.BOOLEAN and field_type is not inferred_field_type:
                logger.warning(
                    "Detected type mismatch on field {field_name}: expected {expected_type} "
                    "type, got {actual_type}".format(
                        field_name=field_name,
                        expected_type=field_type.value,
                        actual_type=inferred_field_type.value,
                    )
                )



from typing import List

from aporia.core.context import get_context
from aporia.core.core_api import safe_api_function
from aporia.core.errors import AporiaError


class ModelVersion:
    """Model version."""

    def __init__(self, id: str, name: str, model_id: str, model_type: str):
        """Initializes a ModelVersion object.

        Args:
            id: Model Version ID.
            name: Model version name (e.g v1).
            model_id: Model identifier, as received from the Aporia dashboard.
            model_type: Model type (also known as objective - see notes).

        Notes:
            * The supported model types are:
                * "regression" - for regression models
                * "binary" - for binary classification models
                * "multiclass" - for multiclass classification models
                * "multi-label" - for multi-label classification models
                * "ranking" - for ranking models
        """
        self.id = id
        self.model_id = model_id
        self.model_type = model_type
        self.name = name

    @staticmethod
    def from_dict(data: dict) -> "ModelVersion":
        """Initializes a ModelVersion object from a Python dictionary.

        Args:
            data: Python dictionary to deserialize.

        Returns:
            ModelVersion object
        """
        return ModelVersion(
            id=data["id"],
            name=data["name"],
            model_id=data["model_id"],
            model_type=data["model_type"],
        )


@safe_api_function("Get model versions failed, error: {}")
def get_model_versions(model_id: str) -> List[ModelVersion]:
    """Retrieve all versions of a model.

    Args:
        model_id: The ID of the model to retrieve versions for.

    Returns:
        List of model versions.
    """
    query = """
        query($modelID: String!) {
            model_versions(where: { model_id: { _eq: $modelID } }) {
                id
                model_id
                model_type
                name
            }
        }
    """

    variables = {
        "modelID": model_id,
    }

    context = get_context()
    response = context.event_loop.run_coroutine(
        context.http_client.graphql(query, variables, graphql_url="/v1/tenant-hasura/v1/graphql")
    )

    return [ModelVersion.from_dict(model_version) for model_version in response["model_versions"]]


@safe_api_function("Deleting model version failed, error: {}")
def delete_model_version(model_id: str, model_version: str):
    """Delete a model version by name.

    Args:
        model_id: Model identifier, as received from the Aporia dashboard.
        model_version: The name of the model version to delete.
    """
    model_versions = get_model_versions(model_id)
    if len(model_versions) == 1:
        raise AporiaError("Cannot delete model version if there is only one version for a model")

    model_version_to_delete = next(
        filter(lambda item: item.name == model_version, model_versions), None
    )
    if model_version_to_delete is None:
        raise AporiaError(f"Couldn't find model version: {model_version}")

    context = get_context()

    response = context.event_loop.run_coroutine(
        context.http_client.graphql(
            query="""
                mutation deleteESIndex(
                    $modelId: String,
                    $modelVersionId: String,
                    $environmentId: String,
                    $dataTimestamp: DateTime,
                    $indexType: String
                ) {
                    deleteESIndex(
                        modelId: $modelId,
                        modelVersionId: $modelVersionId,
                        environmentId: $environmentId,
                        dataTimestamp: $dataTimestamp,
                        indexType: $indexType
                    ) {
                        acknowledged
                    }
                }
            """,
            variables={"modelVersionId": model_version_to_delete.id, "indexType": None},
            graphql_url="/v1/tenant-hasura/v1/graphql",
        )
    )

    if not response["deleteESIndex"].get("acknowledged"):
        raise AporiaError(
            f"Could not delete data for model version: {model_version}. This might be a temporary issue. Please try again later."
        )

    response = context.event_loop.run_coroutine(
        context.http_client.graphql(
            query="""
                mutation deleteModelVersion($modelVersionId: UUID!) {
                    deleteModelVersion(modelVersionId: $modelVersionId) {
                        deleted
                    }
                }
            """,
            variables={"modelVersionId": model_version_to_delete.id},
            graphql_url="/v1/tenant-hasura/v1/graphql",
        )
    )

    if not response["deleteModelVersion"]["deleted"]:
        raise AporiaError(f"Failed to delete model version: {model_version}")



import base64
from typing import Any, Dict

from aporia.core.http_client import HttpClient
from aporia.inference.types.model_artifact import ModelArtifactType


async def upload_model_artifact(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    model_artifact: bytes,
    artifact_type: ModelArtifactType,
):
    """Logs fitted model object.

    Args:
        http_client: Http client
        model_id: Model ID.
        model_version: Model version.
        model_artifact: Binary model artifact
        artifact_type: Model artifact type
    """
    artifact_base64 = base64.b64encode(model_artifact).decode("ascii")
    await http_client.post(
        url=f"/models/{model_id}/versions/{model_version}/model_artifact",
        data={"artifact_base64": artifact_base64, "artifact_type": artifact_type.value},
    )


async def log_index_to_word_mapping(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    index_to_word_mapping: Dict[Any, Any],
):
    """Logs index to word mapping.

    Args:
        http_client: Http client
        model_id: Model ID.
        model_version: Model version.
        index_to_word_mapping: A mapping between a numeric index to a word.
    """
    mapping_list = list(index_to_word_mapping.items())
    try:
        int(mapping_list[0][0])  # If there's no error after this line, the key is a number
    except ValueError:
        # If the keys are not int, we need to flip the dictionary so that they will be
        mapping_list = [(value, key) for key, value in mapping_list]

    index_to_word_mapping_before_serialization = {str(key): value for key, value in mapping_list}

    await http_client.post(
        url=f"/models/{model_id}/versions/{model_version}/index_to_word_mapping",
        data={"index_to_word_mapping": index_to_word_mapping_before_serialization},
    )


async def set_feature_importance(
    http_client: HttpClient, model_id: str, model_version: str, feature_importance: Dict[str, float]
):
    """Logs fitted model object.

    Args:
        http_client: Http client
        model_id: Model ID.
        model_version: Model version.
        feature_importance: Feature name to importance mapping.
    """
    await http_client.post(
        url=f"/models/{model_id}/versions/{model_version}/feature_importance",
        data={"feature_importance": feature_importance},
    )



from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

from aporia.core.http_client import HttpClient


class DataSource(ABC):
    """General data source representation."""

    @property
    @abstractmethod
    def _type(self) -> str:
        raise ValueError("Data source type is not defined")

    @property
    @abstractmethod
    def _config(self) -> Dict[str, Any]:
        raise ValueError("Data source config is not defined")


class SparkSQLDataSource(DataSource):
    """Spark SQL data source repesentation."""

    def __init__(
        self,
        database: str,
        query: str,
        livy_url: str,
        spark_config: Optional[Dict[str, str]] = None,
        sample: float = 1.0,
        batch_size: int = 100000000,
    ):
        """Spark SQL data source properties initialization."""
        self.database = database
        self.query = query
        self.livy_url = livy_url
        self.spark_config = spark_config
        self.sample = sample
        self.batch_size = batch_size

    @property
    def _type(self) -> str:
        return "spark-sql"

    @property
    def _config(self) -> Dict[str, Any]:
        return {
            "database": self.database,
            "query": self.query,
            "livy_url": self.livy_url,
            "spark_config": self.spark_config,
            "sample": self.sample,
            "batch_size": self.batch_size,
        }


def _get_query_data(
    dataset: str,
    environment: str,
    data_source: DataSource,
    id_column: str,
    timestamp_column: str,
    predictions: Optional[Dict[str, str]] = None,
    features: Optional[Dict[str, str]] = None,
    labels: Optional[Dict[str, str]] = None,
    raw_inputs: Optional[Dict[str, str]] = None,
) -> Dict[str, Any]:
    data = {
        "name": dataset,
        "data_source": {"type": data_source._type, "config": data_source._config},
        "sync": {
            "enabled": True,
            "environment": environment,
        },
        "id_column": id_column,
        "timestamp_column": timestamp_column,
    }

    if predictions is not None:
        data["predictions"] = predictions

    if features is not None:
        data["features"] = features

    if labels is not None:
        data["labels"] = labels

    if raw_inputs is not None:
        data["raw_inputs"] = raw_inputs

    return data


async def connect_dataset(
    http_client: HttpClient,
    model_id: str,
    model_version: str,
    dataset: str,
    environment: str,
    data_source: DataSource,
    id_column: str,
    timestamp_column: str,
    predictions: Optional[Dict[str, str]] = None,
    features: Optional[Dict[str, str]] = None,
    labels: Optional[Dict[str, str]] = None,
    raw_inputs: Optional[Dict[str, str]] = None,
):
    """Connect to a dataset.

    Args:
        http_client: HTTP client.
        model_id: Model id.
        model_version: Model version.
        dataset: Dataset name.
        environment: Environment.
        data_source: Data source object (for example, `SparkSQLDataSource`).
        id_column: Name of the ID column.
        timestamp_column: Name of the timestamp column.
        predictions: Predictions -> columns mappping. Defaults to None.
        features: Features -> columns mappping. Defaults to None.
        labels: Labels -> columns mappping. Defaults to None.
        raw_inputs: Raw inputs -> columns mappping. Defaults to None.
    """
    await http_client.post(
        url=f"/models/{model_id}/versions/{model_version}/datasets",
        data=_get_query_data(
            dataset=dataset,
            environment=environment,
            data_source=data_source,
            id_column=id_column,
            timestamp_column=timestamp_column,
            features=features,
            predictions=predictions,
            labels=labels,
            raw_inputs=raw_inputs,
        ),
        timeout=60 * 10,
    )



from enum import Enum
import logging
from typing import Dict, List, Optional, Union

from numpy import ndarray
from pandas import DataFrame

from aporia.core.errors import AporiaError
from aporia.core.logging_utils import LOGGER_NAME
from aporia.core.types.field import FieldCategory, FieldType
from aporia.core.utils import generate_features_schema_from_shape
from aporia.pandas.pandas_utils import infer_type_from_dtype_and_data

logger = logging.getLogger(LOGGER_NAME)


class TrainingDataset(Enum):
    """Training dataset names."""

    TRAINING = "training"
    TEST = "test"


class TrainingDataType(Enum):
    """Training data types."""

    FEATURES = "features"
    LABELS = "labels"
    PREDICTIONS = "predictions"
    RAW_INPUTS = "raw inputs"

    def to_field_category(self) -> FieldCategory:
        """Converts training data type to corresponding field category."""
        if self is TrainingDataType.FEATURES:
            return FieldCategory.FEATURES
        elif self is TrainingDataType.PREDICTIONS or self is TrainingDataType.LABELS:
            return FieldCategory.PREDICTIONS
        elif self is TrainingDataType.RAW_INPUTS:
            return FieldCategory.RAW_INPUTS

        # This is unreachable, the if-elif cluases above should cover all possible values
        raise AporiaError("Invalid training data type")


class TrainingValidator:
    """Training data validator."""

    def __init__(
        self, dataset_type: TrainingDataset, schema: Dict[FieldCategory, Dict[str, FieldType]]
    ):
        """Initializes a training validator.

        Args:
            dataset_type: Type of dataset (training or test)
            schema: Model version schema
        """
        self.dataset_type = dataset_type
        self.schema = schema

    def validate_field_category_is_in_schema(self, training_data_type: TrainingDataType):
        """Validates that a field category was defined in the model version schema.

        Args:
            training_data_type: Training data type of the dataframe that is being validated.
        """
        if training_data_type.to_field_category() not in self.schema:
            raise AporiaError(
                "Cannot report {} set {}, because it was not defined during model "
                "version creation.".format(self.dataset_type.value, training_data_type.value)
            )

    def validate_no_field_mismatch_in_category(
        self, columns: List[str], training_data_type: TrainingDataType
    ):
        """Validates that no fields (columns) are missing or undefined.

        Args:
            columns: List of columns in the dataframe.
            training_data_type: Training data type of the dataframe that is being validated.
        """
        columns_set = set(columns)
        field_category = training_data_type.to_field_category()

        # Check for missing fields
        missing_fields = self.schema[field_category].keys() - columns_set
        if len(missing_fields) > 0:
            logger.warning(
                "Detected missing {training_type} in the {dataset} data - the following "
                "{training_type} were defined during model version creation but are "
                "not present in the {dataset} data: {missing_fields}".format(
                    training_type=training_data_type.value,
                    dataset=self.dataset_type.value,
                    missing_fields=missing_fields,
                )
            )

        # Check for undefined fields
        undefined_fields = columns_set - self.schema[field_category].keys()
        if len(undefined_fields) > 0:
            logger.warning(
                "Detected undefined {training_type} in the {dataset} data - the following "
                "{training_type} were found in the {dataset} data, but were not defined "
                "during model version creation: {undefined_fields}".format(
                    training_type=training_data_type.value,
                    dataset=self.dataset_type.value,
                    undefined_fields=undefined_fields,
                )
            )

    def _validate_field_types_match_schema(self, field_category: FieldCategory, data: DataFrame):
        missing_fields = self.schema[field_category].keys() - set(data.columns)

        for field_name, field_type in self.schema[field_category].items():
            if field_name in missing_fields:
                continue

            column_data_without_nulls = data[field_name].dropna().infer_objects()
            inferred_field_type = infer_type_from_dtype_and_data(
                dtype=column_data_without_nulls.dtype, data=column_data_without_nulls  # type: ignore
            )

            # Don't verify boolean fields, since we're going to cast them anyway
            if field_type != FieldType.BOOLEAN and field_type != inferred_field_type:
                logger.warning(
                    "Detected type mismatch on field {field_name}: expected {expected_type} "
                    "type, got {actual_type}".format(
                        field_name=field_name,
                        expected_type=field_type.value,
                        actual_type=column_data_without_nulls.dtype,
                    )
                )

    def _validate_feature_count(self, features: ndarray):
        feature_count = len(generate_features_schema_from_shape(features.shape[1:]))
        expected_feature_count = len(self.schema[FieldCategory.FEATURES])
        if feature_count != expected_feature_count:
            raise AporiaError(
                f"Detected length mismatch on {self.dataset_type.value} set features ndarray, "
                f"expected {expected_feature_count}, got {feature_count}."
            )

    def validate(
        self,
        features: Union[DataFrame, ndarray],
        labels: DataFrame,
        predictions: Optional[DataFrame] = None,
        raw_inputs: Optional[DataFrame] = None,
    ):
        """Validates training dataframes.

        Args:
            features: Features dataframe or ndarray
            labels: Labels (predictions) dataframe
            predictions: Predictions dataframe
            raw_inputs: Raw inputs dataframe
        """
        tabular_features = None
        if isinstance(features, DataFrame):
            tabular_features = features
        else:
            self._validate_feature_count(features)

        for training_data_type, data in [
            (TrainingDataType.FEATURES, tabular_features),
            (TrainingDataType.LABELS, labels),
            (TrainingDataType.PREDICTIONS, predictions),
            (TrainingDataType.RAW_INPUTS, raw_inputs),
        ]:
            if data is None:
                continue

            self.validate_field_category_is_in_schema(training_data_type)
            self.validate_no_field_mismatch_in_category(
                columns=data.columns, training_data_type=training_data_type
            )
            self._validate_field_types_match_schema(
                field_category=training_data_type.to_field_category(), data=data
            )



from .pyspark_inference_model import PySparkInferenceModel
from .pyspark_training_model import PySparkTrainingModel
from .pyspark_utils import infer_schema_from_pyspark_dataframe

__all__ = ["PySparkInferenceModel", "PySparkTrainingModel", "infer_schema_from_pyspark_dataframe"]
